/*
 * $Id: stickedenemy.d,v 1.1.1.1 2004/04/03 10:36:32 kenta Exp $
 *
 * Copyright 2004 Kenta Cho. All rights reserved.
 */
module abagames.tf.stickedenemy;

private:
import std.math;
import abagames.util.actor;
import abagames.util.actorpool;
import abagames.util.vector;
import abagames.tf.tumikiset;
import abagames.tf.ship;
import abagames.tf.enemy;
import abagames.tf.bulletactorpool;
import abagames.tf.bullettarget;
import abagames.tf.bulletinst;
import abagames.tf.field;
import abagames.tf.fragment;
import abagames.tf.gamemanager;
import abagames.tf.soundmanager;

/**
 * Enemies that are sticked to my ship.
 */
public class StickedEnemy: Actor {
 private:
  Ship ship;
  Field field;
  GameManager manager;
  TumikiSet tumikiSet;
  Vector ofs;
  float stDeg;
  Vector pos;
  float deg;
  Vector[4] colDatums;
  bool isMyShip;
  static const int TOP_BULLET_MAX = 16;
  EnemyTopBullet[TOP_BULLET_MAX] topBullet;
  int topBulletNum;
  int barragePtnIdx;
  BulletActorPool bullets;
  ActorPool fragments;
  VirtualBulletTarget target;
  int cnt;

  public override Actor newActor() {
    return new StickedEnemy;
  }

  public override void init(ActorInitializer ini) {
    StickedEnemyInitializer sei = (StickedEnemyInitializer) ini;
    ship = sei.ship;
    field = sei.field;
    bullets = sei.bullets;
    fragments = sei.fragments;
    manager = sei.manager;
    ofs = new Vector;
    pos = new Vector;
    target = new VirtualBulletTarget;
    foreach (inout Vector cd; colDatums)
      cd = new Vector;
    foreach (inout EnemyTopBullet etb; topBullet)
      etb = new EnemyTopBullet;
  }

  public void set(float x, float y, float d, TumikiSet ts, int bpi) {
    ofs.x = x; ofs.y = y;
    stDeg = d;
    tumikiSet = ts;
    barragePtnIdx = bpi;
    isMyShip = false;
    addTopBullets();
    cnt = 0;
    isExist = true;
  }

  private void addTopBullets() {
    topBulletNum = tumikiSet.addTopBullets
      (barragePtnIdx, bullets, topBullet, target, BulletInst.Type.SHIP);
  }

  private void removeTopBullets() {
    for (int i = 0 ; i < topBulletNum; i++)
      if (topBullet[i].actor)
	topBullet[i].actor.removeForced();
  }

  public void remove() {
    removeTopBullets();
    isExist = false;
  }

  private void setTopBulletsPos() {
    for (int i = 0 ; i < topBulletNum; i++) {
      EnemyTopBullet etb = topBullet[i];
      if (etb.actor) {
	float stox = etb.tumiki.ofs.x * cos(deg) - etb.tumiki.ofs.y * sin(deg);
	float stoy = etb.tumiki.ofs.x * sin(deg) + etb.tumiki.ofs.y * cos(deg);
	etb.actor.bullet.pos.x = pos.x + stox;
	etb.actor.bullet.pos.y = pos.y + stoy;
	etb.actor.bullet.deg = deg - PI / 2;
      }
    }
  }

  public void setAsMyShip(TumikiSet ts) {
    ofs.x = ofs.y = stDeg = 0;
    tumikiSet = ts;
    isMyShip = true;
    isExist = true;
  }

  public void breakIntoFragments() {
    tumikiSet.breakIntoFragments(fragments, pos, deg);
  }

  private const float COLLISION_RATIO = 0.8;

  public override void move() {
    deg = stDeg + ship.deg;
    float sd = sin(deg) * COLLISION_RATIO, cd = cos(deg) * COLLISION_RATIO;
    pos.x = ship.pos.x + ofs.x * cos(ship.deg) - ofs.y * sin(ship.deg);
    pos.y = ship.pos.y + ofs.x * sin(ship.deg) + ofs.y * cos(ship.deg);
    colDatums[0].x = pos.x + tumikiSet.sizeXm * cd;
    colDatums[0].y = pos.y + tumikiSet.sizeXm * sd;
    colDatums[1].x = pos.x - tumikiSet.sizeYm * sd;
    colDatums[1].y = pos.y + tumikiSet.sizeYm * cd;
    colDatums[2].x = pos.x + tumikiSet.sizeXp * cd;
    colDatums[2].y = pos.y + tumikiSet.sizeXp * sd;
    colDatums[3].x = pos.x - tumikiSet.sizeYp * sd;
    colDatums[3].y = pos.y + tumikiSet.sizeYp * cd;
    target.pos.x = pos.x - cos(deg) * Ship.TARGET_DISTANCE;
    target.pos.y = pos.y - sin(deg) * Ship.TARGET_DISTANCE;
    setTopBulletsPos();
    cnt++;
    if (tumikiSet.fireScoreInterval > 0 &&
	(cnt % tumikiSet.fireScoreInterval) == 0 &&
	!field.checkHit(pos)) {
      manager.addScore(tumikiSet.fireScore, pos);
      SoundManager.playSe(SoundManager.Se.STICKED_BONUS);
    }
  }

  public override void draw() {
    if (!isMyShip)
      tumikiSet.draw(pos, 0.2, deg);
  }

  public bool checkHit(Vector pos) {
    if (pos.checkSide(colDatums[0], colDatums[1]) * 
	pos.checkSide(colDatums[3], colDatums[2]) < 0 &&
	pos.checkSide(colDatums[1], colDatums[2]) * 
	pos.checkSide(colDatums[0], colDatums[3]) < 0)
      return true;
    else
      return false;
  }
}

public class StickedEnemyInitializer: ActorInitializer {
 private:
  Ship ship;
  Field field;
  BulletActorPool bullets;
  ActorPool fragments;
  GameManager manager;

  public this(Ship ship, Field field, BulletActorPool bullets, ActorPool fragments,
	      GameManager manager) {
    this.ship = ship;
    this.field = field;
    this.bullets = bullets;
    this.fragments = fragments;
    this.manager = manager;
  }
}

public class StickedEnemyPool: ActorPool {
 private:

  public this(int n, ActorInitializer ini) {
    auto StickedEnemy seClass = new StickedEnemy;
    super(n, seClass, ini);
  }

  public bool checkHit(Vector pos) {
    foreach (Actor ac; actor) {
      if (ac.isExist) {
	StickedEnemy se = (StickedEnemy) ac;
	if (se.checkHit(pos))
	  return true;
      }
    }
    return false;
  }

  public bool checkHitWithoutMyShip(Vector pos) {
    foreach (Actor ac; actor) {
      if (ac.isExist) {
	StickedEnemy se = (StickedEnemy) ac;
	if (!se.isMyShip && se.checkHit(pos))
	  return true;
      }
    }
    return false;
  }

  public void removeAllEnemies() {
    foreach (Actor ac; actor) {
      if (ac.isExist) {
	StickedEnemy se = (StickedEnemy) ac;
	if (!se.isMyShip) {
	  se.breakIntoFragments();
	  se.remove();
	}
      }
    }
  }
}
